#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\api\serial\__init__.py
# @Time    :   2022-05-03 03:34:14
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

__all__ = ["Serializer", "Deserializer"]

from .serializer import Serializer
from .deserializer import Deserializer



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\model\optimizer\__init__.py
# @Time    :   2022-05-03 16:07:47
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

import torch
import torch.nn as nn
import torch.optim as optim

optimizer_factory = {
    "sgd": optim.SGD,
    "adam": optim.Adam,
    "adagrad": optim.Adagrad,
}


def build_optimizer(optim_name: str):
    if optim_name not in optimizer_factory.keys():
        raise ValueError(f"Unknown optimizer name: {optim_name}")

    optimizer = optimizer_factory[optim_name]

    return optimizer



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\model\nn\cv\__init__.py
# @Time    :   2022-07-16 23:32:13
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

from .resnet import ResNet
from .resnet_mnist import ResNetMNIST
from .mlp import MLP
from .alexnet_cifar10 import AlexNetCIFAR10
from .cnn_cifar10 import CNN2CIFAR10, CNN4CIFAR10
from .cnn_mnist import CNNMNIST
from .densenet import DenseNet
from .unet import UNet, UNetMini


cv_model_factory = {
    "resnet": ResNet,
    "resnet18": ResNet,
    "resnet_mnist": ResNetMNIST,
    "mlp": MLP,
    "alexnet_cifar10": AlexNetCIFAR10,
    "cnn_cifar10": CNN2CIFAR10,
    "cnn2_cifar10": CNN2CIFAR10,
    "cnn4_cifar10": CNN4CIFAR10,
    "cnn_mnist": CNNMNIST,
    "densenet": DenseNet,
    "unet": UNet,
    "unet_mini": UNetMini,
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\core\register.py
# @Time    :   2022-07-16 23:53:15
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0


from fedhf.api import dp_mechanism_factory, dp_clip_factory
from fedhf.algor.base import algor_factory
from fedhf.component import (
    aggregator_factory,
    attactor_factory,
    encryptor_factory,
    evaluator_factory,
    sampler_factory,
    selector_factory,
    trainer_factory,
)
from fedhf.model import (
    model_factory,
    optimizer_factory,
    criterion_factory,
    lr_schedule_factory,
)
from fedhf.dataset import dataset_factory
from fedhf.core import coordinator_factory, server_factory, client_factory

components = {
    "dp_mechanism": dp_mechanism_factory,
    "dp_clip": dp_clip_factory,
    "algor": algor_factory,
    "aggregator": aggregator_factory,
    "agg": aggregator_factory,  # alias
    "attactor": attactor_factory,
    "encryptor": encryptor_factory,
    "evaluator": evaluator_factory,
    "sampler": sampler_factory,
    "selector": selector_factory,
    "trainer": trainer_factory,
    "model": model_factory,
    "optimizer": optimizer_factory,
    "criterion": criterion_factory,
    "loss": criterion_factory,  # alias
    "lr_schedule": lr_schedule_factory,
    "dataset": dataset_factory,
    "coordinator": coordinator_factory,
    "server": server_factory,
    "client": client_factory,
}


def register(component_name: str, component_dict: dict):
    if component_name not in components.keys():
        raise ValueError(f"Unknown component name: {component_name}")
    components[component_name].update(component_dict)


def register_all(component_dict: dict):
    for component_name, component_dict in component_dict.items():
        register(component_name, component_dict)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\component\__init__.py
# @Time    :   2022-05-03 16:01:31
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

from .aggregator import *
from .attactor import *
from .encryptor import *
from .evaluator import *
from .sampler import *
from .selector import *
from .trainer import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\algor\async_\fedasync\__init__.py
# @Time    :   2022-07-15 12:54:17
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

import os
from fedhf.api import Config
from .fedasync_aggregator import FedAsyncAggregator
from .fedasync_trainer import FedAsyncTrainer

components = {
    "agg": {"fedasync": FedAsyncAggregator},
    "trainer": {"fedasync": FedAsyncTrainer},
}

default_params = Config()
default_params.load(os.path.join(os.path.dirname(__file__), "default_params.yaml"))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\api\dpm\sensitivity.py
# @Time    :   2022-05-02 22:19:28
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

import numpy as np


def calculate_sensitivity(lr, clip, data_size):
    """
    Calculate the sensitivity of the privacy mechanism.
    :param lr: learning rate
    :param clip: clipping bound
    :param data_size: data size
    :return: the sensitivity of the privacy mechanism
    """
    return 2 * lr * clip / data_size



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\algor\base.py
# @Time    :   2022-07-15 21:34:03
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0


from .async_ import fedasync
from .sync import fedavg, fedprox

algor_factory = {
    "fedasync": fedasync,
    "fedavg": fedavg,
    "fedprox": fedprox,
}


def build_algor(algor_type: str):
    if algor_type not in algor_factory.keys():
        raise ValueError(f"{algor_type} is not a valid algor name")

    return algor_factory[algor_type]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\algor\async_\fedasync\fedasync_aggregator.py
# @Time    :   2022-07-15 13:17:23
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0


import time
import torch
import torch.nn as nn
from ....component.aggregator.async_aggregator import AsyncAggregator


class FedAsyncAggregator(AsyncAggregator):
    def __init__(self, args) -> None:
        super(FedAsyncAggregator, self).__init__(args)

        self.algor = args.algor
        assert self.algor != "" and self.algor is not None

        self.stragegy = args[self.algor].strategy
        assert self.stragegy != "" and self.stragegy in [
            "constant",
            "hinge",
            "polynomial",
        ]

        self.a = args[self.algor].a if args[self.algor].get("a") is not None else None
        self.b = args[self.algor].b if args[self.algor].get("b") is not None else None
        self.alpha = (
            args[self.algor].alpha
            if args[self.algor].get("alpha") is not None
            else None
        )

        # asserts
        assert self.alpha is not None, "alpha is required"
        if self.stragegy == "hinge":
            assert self.a is not None, "a is required for strategy hinge"
            assert self.b is not None, "b is required for strategy hinge"
        elif self.stragegy == "polynomial":
            assert self.a is not None, "a is required for strategy polynomial"

    def agg(self, server_param: torch.Tensor, client_param: torch.Tensor, **kwargs):
        if not self._check_agg():
            return

        # model version is required for staleness
        if "server_model_version" not in kwargs.keys():
            raise ValueError("Missing key: server_model_version")
        if "client_model_version" not in kwargs.keys():
            raise ValueError("Missing key: client_model_version")

        server_model_version = kwargs["server_model_version"]
        client_model_version = kwargs["client_model_version"]

        if server_model_version < client_model_version:
            raise ValueError(
                f"server_model_version {server_model_version} < client_model_version {client_model_version}"
            )

        alpha = self._get_alpha(staleness=server_model_version - client_model_version)
        new_param = torch.mul(1 - alpha, server_param) + torch.mul(alpha, client_param)

        # assert torch.equal(new_param, server_param) == False
        # assert torch.equal(new_param, client_param) == False

        self.logger.info(
            f"Aggregated server model version: {server_model_version}, client model version: {client_model_version}"
        )
        self.logger.info(
            f"FedAsyncAggregator agg alpha: {alpha} with stragegy: {self.stragegy}"
        )

        result = {
            "param": new_param,
            "model_version": server_model_version + 1,
            "model_time": time.time(),
        }
        return result

    def _get_alpha(self, staleness: int):
        if self.stragegy == "constant":
            return torch.mul(self.alpha, 1)
        elif self.stragegy == "hinge" and self.b is not None and self.a is not None:
            if staleness <= self.b:
                return torch.mul(self.alpha, 1)
            else:
                return torch.mul(self.alpha, 1 / (self.a * ((staleness - self.b) + 1)))
        elif self.stragegy == "polynomial" and self.a is not None:
            return torch.mul(self.alpha, (staleness + 1) ** (-self.a))
        else:
            raise ValueError("Unknown strategy: {}".format(self.stragegy))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\api\logger\logger.py
# @Time    :   2022-05-03 15:58:38
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

import os
import logging
import sys
import time
from typing import Dict, Optional, Union

from .base_logger import BaseLogger, logger_map

logging.basicConfig(
    stream=sys.stdout,
    level=logging.INFO,
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
)


class Logger(BaseLogger):
    class __Logger(BaseLogger):
        def __init__(self, args):
            if args.log_level in logger_map:
                self.log_level = logger_map[args.log_level]
            else:
                raise "No such log level!"

            if args.log_name is not None:
                self.log_name = args.log_name
            else:
                self.log_name = "root"

            self.logger = logging.getLogger(self.log_name)
            self.logger.setLevel(self.log_level)

            formatter = logging.Formatter(
                "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
            )

            self.log_metric = args.log_metric

            file_handler = logging.FileHandler(args.log_file, mode="w")
            file_handler.setLevel(level=self.log_level)
            file_handler.setFormatter(formatter)
            self.logger.addHandler(file_handler)

            # use streamHandler to print to stdout
            stream_handler = logging.StreamHandler(sys.stdout)
            stream_handler.setLevel(level=self.log_level)
            stream_handler.setFormatter(formatter)
            self.logger.addHandler(stream_handler)

            if args.use_wandb:
                import wandb

                self.use_wandb = args.use_wandb
                wandb.init(
                    project=args.project_name,
                    config=args,
                    reinit=args.wandb_reinit,
                    name=args.name,
                )

        def debug(self, log_str: str) -> None:
            self.logger.debug(log_str)

        def info(self, log_str: str) -> None:
            self.logger.info(log_str)

        def warning(self, log_str: str) -> None:
            self.logger.warning(log_str)

        def error(self, log_str: str) -> None:
            self.logger.error(log_str)

        def log(self, log_dict: dict, *args, **kwargs) -> None:
            # log one line in result.csv
            with open(self.log_file, "a") as f:
                f.write(str(log_dict) + "\n")

            if self.use_wandb:
                self.to_wandb(log_dict, *args, **kwargs)

        def to_wandb(self, log_dict: dict, *args, **kwargs) -> None:
            import wandb

            wandb.log(log_dict, *args, **kwargs)

        def log_metric(self, log_info: Union[Dict, str] = None) -> None:
            # log one line in result.csv
            with open(self.log_metric, "a") as f:
                f.write(str(log_info) + "\n")

    __instance = None

    def __new__(cls, args):
        if not cls.__instance:
            cls.__instance = Logger.__Logger(args)
        return cls.__instance

    def debug(self, log_str: str) -> None:
        self.__instance.debug(log_str)

    def info(self, log_str: str) -> None:
        self.__instance.info(log_str)

    def warning(self, log_str: str) -> None:
        self.__instance.warning(log_str)

    def error(self, log_str: str) -> None:
        self.__instance.error(log_str)

    def log(self, log_dict, *args, **kwargs) -> None:
        self.__instance.log(log_dict, args, kwargs)

    def log_metric(self, log_info, *args, **kwargs) -> None:
        self.__instance.log_metric(log_info, args, kwargs)

    def to_wandb(self, log_dict, *args, **kwargs) -> None:
        if self.__instance.use_wandb:
            self.__instance.to_wandb(log_dict, args, kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\api\metric\metric.py
# @Time    :   2022-03-19 17:50:33
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

from typing import Dict, List
import json
import warnings


class Metric(object):
    def __init__(self, metric: Dict):
        self.metric = metric
        for _key in ["round", "epoch", "id"]:
            if _key not in self.metric:
                self.metric[_key] = 0
                warnings.warn(f"Metric {_key} not found")

    def get_metric(self, name: str):
        if name in self.metric:
            return self.metric[name]
        raise KeyError(f"Metric {name} not found")

    def get_metrics(self) -> Dict:
        return self.metric

    def update(self, metric: Dict):
        self.metric.update(metric)

    def to_json(self) -> str:
        return json.dumps(self.metric)


class Metrics(object):
    def __init__(self, metrics: List[Metric]):
        self.metrics = metrics

    def add(self, metric: Metric):
        self.metrics.append(metric)

    def get_metrics(self) -> List[Metric]:
        return self.metrics

    def save(self, path: str):
        with open(path, "w") as f:
            f.write(json.dumps(self.metrics))

    def load(self, path: str):
        with open(path, "r") as f:
            self.metrics = json.loads(f.read())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\model\nn\cnn_cifar10.py
# @Time    :   2022-05-03 16:07:12
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

import torch
import torch.nn as nn
from torch.nn.modules.activation import Softmax

from ..base_model import BaseModel


class CNN4CIFAR10(BaseModel):
    """
    Implentation of the cnn described in the fedasync
    """

    def __init__(self, args, model_time=None, model_version=0):
        super().__init__(args, model_time, model_version)

        self.num_classes = args.num_classes

        self.net = nn.Sequential(
            nn.Conv2d(3, 64, kernel_size=3, padding=1),
            nn.LeakyReLU(0.1),
            nn.BatchNorm2d(64),
            nn.Conv2d(64, 64, kernel_size=3, padding=1),
            nn.LeakyReLU(0.1),
            nn.BatchNorm2d(64),
            nn.MaxPool2d(2),
            nn.Dropout(0.25),
            nn.Conv2d(64, 128, kernel_size=3, padding=1),
            nn.LeakyReLU(0.1),
            nn.BatchNorm2d(128),
            nn.Conv2d(128, 128, kernel_size=3, padding=1),
            nn.LeakyReLU(0.1),
            nn.BatchNorm2d(128),
            nn.MaxPool2d(2),
            nn.Dropout(0.25),
            nn.Flatten(),
            nn.Linear(128 * 8 * 8, 512),
            nn.ReLU(0.1),
            nn.Dropout(0.25),
            nn.Linear(512, self.num_classes),
            nn.Softmax(dim=1),
        )

    def forward(self, x):
        return self.net(x)


class CNN2CIFAR10(BaseModel):
    """
    Implentation of the cnn described in the fedasync
    """

    def __init__(self, args, model_time=None, model_version=0):
        super().__init__(args, model_time, model_version)

        self.num_classes = args.num_classes

        self.net = nn.Sequential(
            nn.Conv2d(3, 32, kernel_size=3, padding=1),
            nn.ReLU(),
            nn.MaxPool2d(2),
            nn.Conv2d(32, 64, kernel_size=3, padding=1),
            nn.ReLU(),
            nn.MaxPool2d(2),
            nn.Flatten(),
            nn.Linear(64 * 8 * 8, 128),
            nn.ReLU(),
            nn.Linear(128, self.num_classes),
            nn.Softmax(dim=1),
        )

    def forward(self, x):
        return self.net(x)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\api\utils\__init__.py
# @Time    :   2022-07-15 17:09:56
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\component\attactor\__init__.py
# @Time    :   2022-07-15 13:00:23
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

from .base_attactor import BaseAttactor

attactor_factory = {
    "base_attactor": BaseAttactor,
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\core\coordinator\simulated_async_coordinator.py
# @Time    :   2022-05-03 16:02:37
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

from copy import deepcopy
import os

import numpy as np

from fedhf.core import build_client

from .base_coordinator import SimulatedBaseCoordinator


class SimulatedAsyncCoordinator(SimulatedBaseCoordinator):
    """Simulated Coordinator
    In simulated scheme, the data and model belong to coordinator and there is no need communicator.
    Also, there is no need to instantiate every client.
    """

    def __init__(self, args) -> None:
        super(SimulatedAsyncCoordinator, self).__init__(args)

    def prepare(self) -> None:
        super(SimulatedAsyncCoordinator, self).prepare()

        assert self.args.deploy_mode == "simulated"
        self._model_queue = []
        self._last_update_time = {client_id: 0 for client_id in self.client_list}

    def main(self) -> None:
        try:
            self.max_staleness = self.args[self.args.algor].get("max_staleness", 16)
            self._model_queue.append(deepcopy(self.server.model))

            for i in range(self.args.num_rounds):
                # self.logger.info(f'{self.server.model.get_model_version()}')

                selected_clients = self.server.select(self.client_list)

                self.logger.info(f"Round {i} Selected clients: {selected_clients}")

                for client_id in selected_clients:
                    client = build_client(self.args.deploy_mode)(
                        self.args, client_id, data_size=len(self.data[client_id])
                    )

                    staleness = np.random.randint(
                        low=1,
                        high=min(
                            self.max_staleness,
                            max(self.server.model.get_model_version(), 0) + 1,
                            self.server.model.get_model_version()
                            - self._last_update_time[client_id]
                            + 1,
                        )
                        + 1,
                    )

                    assert (
                        staleness <= max(0, self.server.model.get_model_version()) + 1
                    )
                    assert staleness <= len(self._model_queue)
                    assert (
                        staleness
                        <= self.server.model.get_model_version()
                        - self._last_update_time[client_id]
                        + 1
                    )

                    self.logger.info(
                        f"Client {client_id} staleness: {staleness} start train from model version : {self._model_queue[-staleness].get_model_version()}"
                    )

                    model, result = client.train(
                        data=self.data[client_id],
                        model=deepcopy(self._model_queue[-staleness]),
                    )

                    self.server.update(
                        model,
                        server_model_version=max(
                            0, self.server.model.get_model_version()
                        ),
                        client_model_version=max(0, model.get_model_version()),
                    )

                    result = self.server.evaluate(self.dataset.testset)
                    self.logger.info(
                        f"Server model version {self.server.model.get_model_version()} result: {result}"
                    )
                    if (
                        self.server.model.get_model_version() % self.args.chkp_interval
                        == 0
                    ):
                        self.logger.info(
                            f"Save model: {self.args.exp_name}-{self.server.model.get_model_version()}.pth"
                        )
                        self.server.model.save(
                            os.path.join(
                                self.args.save_dir,
                                f"{self.args.exp_name}-{self.server.model.get_model_version()}.pth",
                            )
                        )

                    self._model_queue.append(deepcopy(self.server.model))
                    self._last_update_time[
                        client_id
                    ] = self.server.model.get_model_version()

                    while len(self._model_queue) > self.max_staleness + 1:
                        self._model_queue.pop(0)
            self.logger.info(f"All rounds finished.")

        except KeyboardInterrupt:
            self.server.model.save()
            self.logger.info(f"Interrupted by user.")

    def finish(self) -> None:
        super(SimulatedAsyncCoordinator, self).finish()

    def run(self) -> None:
        self.prepare()
        self.main()
        self.finish()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\core\attactor\base_attactor.py
# @Time    :   2022-05-03 16:02:07
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

from abc import ABC, abstractmethod
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import grad

from fedhf.model import build_criterion, build_optimizer


class AbsAttactor(ABC):
    def __init__(self, args):
        self.args = args

    @abstractmethod
    def attack(self):
        raise NotImplementedError


class BaseAttactor(AbsAttactor):
    def __init__(self, args):
        super(BaseAttactor, self).__init__(args)
        self.crit = build_criterion(self.args.train_loss)
        self.data_size = (
            self.args.batch_size,
            self.args.input_c,
            self.args.image_size,
            self.args.image_size,
        )
        self.label_size = (self.args.batch_size, self.args.num_classes)

    def attack(self, model, origin_grad):
        dummy_data = (
            torch.randn(self.data_size).to(self.args.device).requires_grad_(True)
        )
        dummy_label = (
            torch.randn(self.label_size).to(self.args.device).requires_grad_(True)
        )
        optimizer = torch.optim.LBFGS([dummy_data, dummy_label])

        for iters in range(300):

            def closure():
                optimizer.zero_grad()
                dummy_pred = model(dummy_data)
                dummy_loss = self.crit(dummy_pred, F.softmax(dummy_label, dim=-1))
                dummy_grad = grad(dummy_loss, model.parameters(), create_graph=True)

                grad_diff = 0
                for gx, gy in zip(dummy_grad, origin_grad):
                    grad_diff += ((gx - gy) ** 2).sum()
                grad_diff.backward()
                return grad_diff

            optimizer.step(closure)

        return dummy_data, dummy_label



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\core\attactor\dlg_attactor.py
# @Time    :   2022-05-03 16:02:11
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

from .base_attactor import BaseAttactor


class DLGAttactor(BaseAttactor):
    def __init__(self, args):
        super(DLGAttactor, self).__init__(args)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\core\server\base_server.py
# @Time    :   2022-05-03 15:45:08
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

from abc import ABC
from torch.utils.data.dataloader import DataLoader

from fedhf.api import Logger
from fedhf.component import (
    build_aggregator,
    build_selector,
    build_evaluator,
    build_encryptor,
)
from fedhf.model import build_model


class AbsServer(ABC):
    def __init__(self) -> None:
        super().__init__()

    def update(self):
        raise NotImplementedError

    def select(self):
        raise NotImplementedError

    def evaluate(self):
        raise NotImplementedError


class BaseServer(AbsServer):
    def __init__(self, args) -> None:
        self.args = args

        self.selector = build_selector(self.args.selector)(self.args)
        self.aggregator = build_aggregator(self.args.agg)(self.args)

        self.model = build_model(self.args.model)(self.args, model_version=0)
        self.evaluator = build_evaluator(self.args.evaluator)(self.args)
        if self.args.get("encryptor"):
            self.encryptor = build_encryptor(self.args.encryptor)(self.args)

        self.logger = Logger(self.args)

    def select(self, client_list: list):
        return self.selector.select(client_list)

    def evaluate(self, dataset, **kwargs):
        dataloader = DataLoader(dataset, batch_size=self.args.batch_size, shuffle=False)
        return self.evaluator.evaluate(
            dataloader=dataloader, model=self.model, device=self.args.device, **kwargs
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\component\sampler\__init__.py
# @Time    :   2022-05-03 16:00:35
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

__all__ = [
    "build_sampler",
    "RandomSampler",
    "NonIIDSampler",
    "sampler_factory",
    "BaseSampler",
]

from .base_sampler import BaseSampler

from .random_sampler import RandomSampler
from .noniid_sampler import NonIIDSampler

sampler_factory = {
    "random": RandomSampler,
    "non-iid": NonIIDSampler,
}


def build_sampler(sam_name: str):
    if sam_name not in sampler_factory.keys():
        raise ValueError(f"Sampler {sam_name} not found.")

    sampler = sampler_factory[sam_name]

    return sampler



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\api\opt\cfg.py
# @Time    :   2022-07-04 14:06:07
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

import os
from pathlib import Path
import ezkfg as ez
import numpy as np
import torch

from .opts import opts
from ..path import *

from ezkfg import Config as EConfig


class Config(ez.Config):
    def __init__(self, *args, **kwargs):
        super().__init__()
        self.__parent__ = None
        self.__key__ = None

        # print("args:", args)
        # print("kwargs:", kwargs)
        self.load_args_kwargs(*args, **kwargs)  # load args and kwargs, highest priority
        # print("init called")

    def parse_cfg(self, *args, **kwargs):
        opt = opts().parse()  # default opts
        self.load(opt)

        cfg = None
        if opt.cfg or self.cfg:  # load cfg file
            cfg = self.cfg if self.cfg else opt.cfg
            self.load_from_file(cfg)

        self.load_args_kwargs(*args, **kwargs)  # load args and kwargs, lowest priority

        if self.cfg and cfg != self.cfg:
            self.load_from_file(self.cfg)

        np.random.seed(self.seed)
        torch.manual_seed(self.seed)

        _prj_name = f"{self.scheme}_{self.num_clients}_{self.num_rounds}_{self.num_epochs}_{self.seed}"

        self.prj_name = self.prj_name if self.prj_name else _prj_name

        # make dirs
        self.save_dir = Path(self.save_dir) / self.exp_name
        self.save_dir = increment_path(
            self.save_dir, exist_ok=self.exist_ok, mkdir=True
        )
        if not self.weights_dir:
            self.weights_dir = self.save_dir / "weights"
        else:
            self.weights_dir = self.save_dir / self.weights_dir
        self.weights_dir.mkdir(parents=True, exist_ok=True)
        self.data_dir = Path(self.data_dir)

        # logger
        if self.log_dir:
            self.log_dir = Path(self.log_dir) / self.exp_name
            self.log_dir.mkdir(parents=True, exist_ok=True)
        else:
            self.log_dir = self.save_dir / "log"
            self.log_dir.mkdir(parents=True, exist_ok=True)

        if self.log_file is None:
            self.log_file = self.log_dir / f"{self.exp_name}.log"
        else:
            self.log_file = self.log_dir / self.log_file

        if self.log_metric is None:
            self.log_metric = self.log_dir / f"{self.exp_name}.metric.log"
        else:
            self.log_metric = self.log_dir / self.log_metric

        # device
        self.gpus_str = str(self.gpus)
        self.gpus = [int(gpu) for gpu in self.gpus.split(",")]
        self.gpus = [i for i in range(len(self.gpus))] if self.gpus[0] >= 0 else [-1]
        self.device = torch.device("cuda" if self.gpus[0] >= 0 else "cpu")
        if self.device != "cpu":
            torch.backends.cudnn.benchmark = True

        self.num_workers = max(self.num_workers, 2 * len(self.gpus))

        if self.scheme == "async":
            self.select_ratio = 1.0

        # print(self.get("dp"))
        if self.get("dp"):
            self.dp.epsilon = self.dp.epsilon / (self.select_ratio * self.num_epochs)

        self.num_clients_per_round = (
            self.num_clients_per_round
            if self.num_clients_per_round
            else int(self.num_clients * self.select_ratio)
        )
        self.num_clients_per_round = max(self.num_clients_per_round, 1)

        # if opt.resume and opt.load_model == "":
        #     opt.load_model = os.path.join(opt.save_dir, f"{opt.name}.pth")
        return self



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\component\aggregator\async_aggregator.py
# @Time    :   2022-05-03 15:59:45
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

import time

import torch
import torch.nn as nn

from fedhf import model

from fedhf.api import Logger
from .base_aggregator import BaseAggregator


class AsyncAggregator(BaseAggregator):
    def __init__(self, args) -> None:
        super(AsyncAggregator, self).__init__(args)
        self.alpha = 0.6

    def agg(self, server_param: torch.Tensor, client_param: torch.Tensor, **kwargs):
        if not self._check_agg():
            return

        alpha = self.alpha
        new_param = torch.mul(1 - alpha, server_param) + torch.mul(alpha, client_param)

        result = {
            "param": new_param,
            "model_version": kwargs["server_model_version"] + 1
            if "server_model_version" in kwargs.keys()
            else 0,
            "model_time": time.time(),
        }
        return result

    def _check_agg(self):
        return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\algor\sync\fedavg\__init__.py
# @Time    :   2022-07-15 13:17:37
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

import os
from fedhf import Config
from .fedavg_aggregator import FedAvgAggregator

components = {"agg": {"fedavg": FedAvgAggregator}}

default_params = Config()
default_params.load(os.path.join(os.path.dirname(__file__), "default_params.yaml"))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\core\client\simulated_client.py
# @Time    :   2022-03-19 22:07:10
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

import torch
import torch.nn as nn
from torch.utils.data import DataLoader, Dataset

from fedhf.component import build_encryptor
from .base_client import BaseClient


class SimulatedClient(BaseClient):
    def __init__(self, args, client_id, **kwargs) -> None:
        super(SimulatedClient, self).__init__(args, client_id)
        assert "data_size" in kwargs.keys()
        self.data_size = kwargs["data_size"]

    def train(self, data: Dataset, model: nn.Module, device="cpu", **kwargs):
        dataloader = DataLoader(data, batch_size=self.args.batch_size)

        result = self.trainer.train(
            dataloader=dataloader,
            model=model,
            num_epochs=self.args.num_epochs,
            client_id=self.client_id,
            device=device,
            encryptor=self.encryptor,
            **kwargs,
        )

        model = result["model"]
        result.pop("model")

        if self.encryptor is not None:
            model = self.encryptor.encrypt_model(model)

        # self.logger.info(f'Finish training on client {self.client_id}, train_loss: {train_loss}')
        return model, result

    def evaluate(self, data: Dataset, model: nn.Module, device="cpu", **kwargs):
        dataloader = DataLoader(data, batch_size=self.args.batch_size)

        result = self.evaluator.evaluate(
            dataloader=dataloader, model=model, client_id=self.client_id, device=device
        )
        if "model" in result:
            model = result["model"]
            result.pop("model")

        self.logger.info(
            f"Finish evaluating on client {self.client_id}, result: {result}"
        )
        return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\component\selector\base_selector.py
# @Time    :   2022-05-03 16:00:57
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

from abc import ABC


class BaseSelector(ABC):
    def __init__(self) -> None:
        super().__init__()

    def select(self):
        raise NotImplementedError



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\dataset\cifar10.py
# @Time    :   2022-05-03 16:06:17
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

from torchvision.datasets import CIFAR10
from torchvision.transforms import Compose, ToTensor, Normalize, Resize
import torch


class CIFAR10Dataset(object):
    def __init__(self, args) -> None:
        super().__init__()

        self.args = args
        self.num_classes = 10

        if args.resize:
            self.transform = Compose(
                [
                    Resize([args.image_size, args.image_size]),
                    ToTensor(),
                    Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5)),
                ]
            )
        else:
            self.transform = Compose(
                [ToTensor(), Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5))]
            )

        self.trainset = CIFAR10(
            root=args.data_dir, train=True, download=True, transform=self.transform
        )
        self.testset = CIFAR10(
            root=args.data_dir, train=False, download=True, transform=self.transform
        )

        self.trainset.num_classes = self.num_classes
        self.testset.num_classes = self.num_classes



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\model\nn\mlp.py
# @Time    :   2022-05-03 16:07:23
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

import torch
import torch.nn as nn

from ..base_model import BaseModel


class MLP(BaseModel):
    def __init__(
        self,
        args,
        model_time=None,
        model_version=0,
        input_dim=28 * 28,
        hidden_dim=128,
        output_dim=10,
    ):
        super().__init__(args, model_time, model_version)
        self.layer_input = nn.Linear(input_dim, hidden_dim)
        self.relu = nn.ReLU()
        self.dropout = nn.Dropout()
        self.layer_hidden = nn.Linear(hidden_dim, output_dim)

    def forward(self, x):
        x = x.view(-1, x.shape[1] * x.shape[-2] * x.shape[-1])
        x = self.layer_input(x)
        x = self.dropout(x)
        x = self.relu(x)
        x = self.layer_hidden(x)
        return x



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\model\nn\alexnet_cifar10.py
# @Time    :   2022-05-03 16:07:01
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

import torch
import torch.nn as nn

from ..base_model import BaseModel


class AlexNetCIFAR10(BaseModel):
    def __init__(self, args, model_time=0):
        super().__init__(args, model_time)

        self.args = args
        self.num_classes = args.num_classes

        self.features = nn.Sequential(
            nn.Conv2d(3, 64, kernel_size=3, stride=2, padding=1),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=2),
            nn.Conv2d(64, 192, kernel_size=3, padding=1),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=2),
            nn.Conv2d(192, 384, kernel_size=3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(384, 256, kernel_size=3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(256, 256, kernel_size=3, padding=1),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=2),
        )
        self.classifier = nn.Sequential(
            # nn.Dropout(),
            nn.Linear(256 * 2 * 2, 4096),
            nn.ReLU(inplace=True),
            # nn.Dropout(),
            nn.Linear(4096, 4096),
            nn.ReLU(inplace=True),
            nn.Linear(4096, args.num_classes),
        )

    def forward(self, x):
        x = self.features(x)
        x = x.view(x.size(0), 256 * 2 * 2)
        x = self.classifier(x)
        return x



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\api\serial\serializer.py
# @Time    :   2022-05-03 15:59:29
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

import torch


class Serializer(object):
    """
    Serializer
    """

    @staticmethod
    def serialize_model_grad(model: torch.nn.Module) -> torch.Tensor:
        grads = [param.data.grad.view(-1) for param in model.parameters()]
        grads = torch.cat(grads)
        grads = grads.cpu()
        return grads

    @staticmethod
    def serialize_model(model: torch.nn.Module) -> torch.Tensor:
        params = [param.data.view(-1) for param in model.parameters()]
        params = torch.cat(params)
        params = params.cpu()
        return params



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\api\dpm\__init__.py
# @Time    :   2022-05-02 22:15:36
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

# dpm is a package for differential privacy mechanisms

import numpy as np
from .sensitivity import calculate_sensitivity
from .gaussian import gaussian_noise, gaussian_clip
from .laplace import laplace_noise, laplace_clip

dp_mechanism_factory = {
    "none": None,
    "gaussian": gaussian_noise,
    "laplace": laplace_noise,
}

dp_clip_factory = {
    "none": None,
    "gaussian": gaussian_clip,
    "laplace": laplace_clip,
}


def build_mechanism(mechanism, sensitivity, size, epsilon, **kwargs):
    """
    Build the privacy mechanism.
    :param mechanism: the mechanism name
    :param sensitivity: the sensitivity of the privacy mechanism
    :param size: the size of the noise
    :param epsilon: the privacy parameter
    :param kwargs: other parameters
    :return: the privacy mechanism
    """
    if mechanism == "none":
        return np.zeros(size)
    elif mechanism in dp_mechanism_factory:
        return dp_mechanism_factory[mechanism](sensitivity, size, epsilon, **kwargs)
    else:
        raise ValueError("Unknown mechanism: %s" % mechanism)


def build_clip_grad(mechanism, model, clip, **kwargs):
    """
    Clip the gradients of the model.
    :param mechanism: the mechanism name
    :param model: the model
    :param clip: the clipping bound
    :param kwargs: other parameters
    :return: None
    """
    if mechanism == "none":
        return
    elif mechanism in dp_clip_factory:
        dp_clip_factory[mechanism](model, clip, **kwargs)
    else:
        raise ValueError("Unknown mechanism: %s" % mechanism)


__all__ = [
    "calculate_sensitivity",
    "gaussian_noise",
    "gaussian_clip",
    "laplace_noise",
    "laplace_clip",
    "build_mechanism",
    "build_clip_grad",
    "dp_mechanism_factory",
    "dp_clip_factory",
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\model\nn\unet.py
# @Time    :   2022-02-26 14:04:05

import torch
import torch.nn as nn
import torch.nn.functional as F

from fedhf.api import Config
from ..base_model import BaseModel

"""
    U-Net: Convolutional Networks for Biomedical Image Segmentation
    Olaf Ronneberger, Philipp Fischer, Thomas Brox
    https://arxiv.org/abs/1505.04597

    Code References: https://github.com/milesial/Pytorch-UNet
    License: GNU General Public License v3.0
"""


class DoubleConv(nn.Module):
    def __init__(
        self,
        in_channel,
        out_channel,
        hidden_channel=None,
        kernel_size=3,
        padding=1,
        bias=True,
    ):
        super(DoubleConv, self).__init__()
        if not hidden_channel:
            hidden_channel = out_channel
        self.double_conv = nn.Sequential(
            nn.Conv2d(
                in_channel,
                hidden_channel,
                kernel_size=kernel_size,
                padding=padding,
                bias=bias,
            ),
            nn.BatchNorm2d(hidden_channel),
            nn.ReLU(inplace=True),
            nn.Conv2d(
                hidden_channel,
                out_channel,
                kernel_size=kernel_size,
                padding=padding,
                bias=bias,
            ),
            nn.BatchNorm2d(out_channel),
            nn.ReLU(inplace=True),
        )

    def forward(self, x):
        return self.double_conv(x)


class DownConv(nn.Module):
    def __init__(self, in_channel, out_channel):
        super(DownConv, self).__init__()
        self.down_conv = nn.Sequential(
            nn.MaxPool2d(2),
            DoubleConv(in_channel, out_channel),
        )

    def forward(self, x):
        return self.down_conv(x)


class UpConv(nn.Module):
    def __init__(self, in_channel, out_channel, bilinear=True):
        super(UpConv, self).__init__()

        if bilinear:
            self.up = nn.Upsample(scale_factor=2, mode="bilinear", align_corners=True)
            self.conv = DoubleConv(in_channel, out_channel, in_channel // 2)
        else:
            self.up = nn.ConvTranspose2d(
                in_channel, in_channel // 2, kernel_size=2, stride=2
            )
            self.conv = DoubleConv(in_channel, out_channel)

    def forward(self, x1, x2):
        x1 = self.up(x1)
        # input is CHW
        diffY = x2.size()[2] - x1.size()[2]
        diffX = x2.size()[3] - x1.size()[3]

        x1 = F.pad(x1, [diffX // 2, diffX - diffX // 2, diffY // 2, diffY - diffY // 2])
        # if you have padding issues, see
        # https://github.com/HaiyongJiang/U-Net-Pytorch-Unstructured-Buggy/commit/0e854509c2cea854e247a9c615f175f76fbb2e3a
        # https://github.com/xiaopeng-liao/Pytorch-UNet/commit/8ebac70e633bac59fc22bb5195e513d5832fb3bd
        x = torch.cat([x2, x1], dim=1)
        return self.conv(x)


class OutConv(nn.Module):
    def __init__(self, in_channel, out_channel):
        super(OutConv, self).__init__()
        self.conv = nn.Conv2d(in_channel, out_channel, kernel_size=1)

    def forward(self, x):
        return self.conv(x)


class UNet(BaseModel):

    default_params = Config(unet={"bilinear": None, "n1": None})

    def __init__(self, args, model_time=None, model_version=0):
        super().__init__(args, model_time, model_version)
        self.add_default_args(self.default_params)
        # print("UNet:", self.args.unet)

        self.input_c = self.args.input_c
        self.output_c = self.args.output_c
        self.bilinear = (
            self.args.unet.bilinear if self.args.unet.bilinear is not None else False
        )

        self.n1 = self.args.unet.n1 if self.args.unet.n1 is not None else 64
        self.filter = [self.n1, self.n1 * 2, self.n1 * 4, self.n1 * 8, self.n1 * 16]

        self.inc = DoubleConv(self.input_c, self.filter[0])
        self.down1 = DownConv(self.filter[0], self.filter[1])
        self.down2 = DownConv(self.filter[1], self.filter[2])
        self.down3 = DownConv(self.filter[2], self.filter[3])
        factor = 2 if self.bilinear else 1
        self.down4 = DownConv(self.filter[3], self.filter[4] // factor)
        self.up1 = UpConv(self.filter[4], self.filter[3] // factor, self.bilinear)
        self.up2 = UpConv(self.filter[3], self.filter[2] // factor, self.bilinear)
        self.up3 = UpConv(self.filter[2], self.filter[1] // factor, self.bilinear)
        self.up4 = UpConv(self.filter[1], self.filter[0], self.bilinear)
        self.outc = OutConv(self.filter[0], self.output_c)

        self.init_weights()

    def init_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

    def forward(self, x):
        x1 = self.inc(x)
        x2 = self.down1(x1)
        x3 = self.down2(x2)
        x4 = self.down3(x3)
        x5 = self.down4(x4)
        x = self.up1(x5, x4)
        x = self.up2(x, x3)
        x = self.up3(x, x2)
        x = self.up4(x, x1)
        pred = self.outc(x)
        return pred


class UNetMini(BaseModel):
    default_params = Config(unet={"bilinear": None, "n1": None})

    def __init__(self, args, model_time=None, model_version=0):
        super().__init__(args, model_time, model_version)
        self.add_default_args(self.default_params)

        self.input_c = self.args.input_c
        self.output_c = self.args.output_c
        self.bilinear = (
            self.args.unet.bilinear if self.args.unet.bilinear is not None else False
        )

        self.n1 = self.args.unet.n1 if self.args.unet.n1 is not None else 64
        self.filter = [self.n1, self.n1 * 2, self.n1 * 4]

        factor = 2 if self.bilinear else 1

        self.inc = DoubleConv(self.input_c, self.filter[0])
        self.down1 = DownConv(self.filter[0], self.filter[1])
        self.down2 = DownConv(self.filter[1], self.filter[2] // factor)
        self.up1 = UpConv(self.filter[2], self.filter[1] // factor, self.bilinear)
        self.up2 = UpConv(self.filter[1], self.filter[0], self.bilinear)
        self.outc = OutConv(self.filter[0], self.output_c)

        self.init_weights()

    def init_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

    def forward(self, x):
        x1 = self.inc(x)
        x2 = self.down1(x1)
        x3 = self.down2(x2)
        x = self.up1(x3, x2)
        x = self.up2(x, x1)
        pred = self.outc(x)
        return pred



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\model\scheduler\__init__.py
# @Time    :   2022-02-25 20:44:28
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

import torch
import torch.nn as nn
import torch.optim as optim

lr_schedule_factory = {
    "step": optim.lr_scheduler.StepLR,
    "multi_step": optim.lr_scheduler.MultiStepLR,
    "exponential": optim.lr_scheduler.ExponentialLR,
    "cosine": optim.lr_scheduler.CosineAnnealingLR,
}


def build_lr_scheduler(lr_schedule_name: str):
    if lr_schedule_name not in lr_schedule_factory.keys():
        raise ValueError(f"Unknown lr_schedule name: {lr_schedule_name}")

    lr_scheduler = lr_schedule_factory[lr_schedule_name]
    return lr_scheduler



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\api\opt\__init__.py
# @Time    :   2022-05-03 15:58:49
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

__all__ = ["opts", "Config"]

from .opts import opts
from .cfg import Config



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\component\encryptor\dp_encryptor.py
# @Time    :   2022-05-02 22:55:08
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

import torch
import torch.nn as nn

from .base_encryptor import BaseEncryptor

from fedhf.api import dpm


class DPEncryptor(BaseEncryptor):
    """
    DPEncryptor is the class for DP encryption.
    """

    def __init__(self, args, **kwargs):
        """
        Constructor.
        :param args: the arguments
        """
        super().__init__(args)
        assert kwargs["data_size"] is not None
        self.data_size = kwargs["data_size"]
        assert hasattr(args, "dp"), "The args must have the dp attribute."
        self.sensitivity = dpm.calculate_sensitivity(
            self.args.lr, self.args.dp.clip, self.data_size
        )

    def generate_noise(self, size):
        """
        Encrypt the given data.
        :param x: the data
        :return: the encrypted data
        """
        return dpm.build_mechanism(
            self.args.dp.mechanism,
            self.sensitivity,
            size,
            self.args.dp.epsilon,
            delta=self.args.dp.delta if hasattr(self.args.dp, "delta") else None,
        )

    def clip_grad(self, model):
        """
        Clip the gradient of the given model.
        :param model: the model
        """
        dpm.build_clip_grad(
            self.args.dp.mechanism,
            model,
            self.args.dp.clip,
        )

    def encrypt_model(self, model):
        """
        Encrypt the given model.
        :param model: the model
        :return: the encrypted model
        """
        self.clip_grad(model)

        with torch.no_grad():
            for k, v in model.named_parameters():
                noise = self.generate_noise(v.shape)
                noise = torch.from_numpy(noise).to(self.args.device)
                v.add_(noise)

        return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\model\criterion\__init__.py
# @Time    :   2022-05-03 16:06:51
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

import torch
import torch.nn as nn
import torch.optim as optim

criterion_factory = {
    "l1": nn.L1Loss,
    "mse": nn.MSELoss,
    "ce": nn.CrossEntropyLoss,
    "bce": nn.BCELoss,
}


def build_criterion(criter_name: str):
    if criter_name not in criterion_factory.keys():
        raise ValueError(f"Unknown criterion name: {criter_name}")

    cirter = criterion_factory[criter_name]
    return cirter



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\component\crypor\base_crypor.py
# @Time    :   2022-05-02 22:36:30
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

from abc import ABC, abstractmethod

import numpy as np


class AbsEncryptor(ABC):
    """
    AbsEncryptor is the base class for all crypor mechanisms.
    """

    @abstractmethod
    def __init__(self, args, **kwargs):
        """
        Constructor.
        :param args: the arguments
        """
        self.args = args

    @abstractmethod
    def generate_noise(self):
        """
        Generate noise.
        :return: the noise
        """
        raise NotImplementedError

    @abstractmethod
    def clip_grad(self, model):
        """
        Clip the gradient of the given model.
        :param model: the model
        """
        raise NotImplementedError

    @abstractmethod
    def encrypt_model(self):
        """
        Encrypt the model.
        :return: the encrypted model
        """
        raise NotImplementedError


class BaseEncryptor(AbsEncryptor):
    """
    BaseEncryptor is the base class for all crypor mechanisms.
    """

    def __init__(self, args, **kwargs):
        """
        Constructor.
        :param args: the arguments
        """
        super().__init__(args)

    def generate_noise(self, size):
        """
        Generate noise.
        :param size: the size
        :return: the noise
        """
        return np.zeros(size)

    def clip_grad(self, model):
        """
        Clip the gradient of the given model.
        :param model: the model
        """
        return

    def encrypt_model(self, model):
        """
        Encrypt the given model.
        :param model: the model
        :return: the encrypted model
        """
        return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\api\logger\base_logger.py
# @Time    :   2022-05-03 15:58:34
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

from abc import ABC, abstractmethod

import logging

logger_map = {
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "warning": logging.WARNING,
    "error": logging.ERROR,
}


class BaseLogger(ABC):
    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    def error(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def warning(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def info(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def debug(self) -> None:
        raise NotImplementedError



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\model\nn\__init__.py
# @Time    :   2022-05-03 16:06:57
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

from .base_model import BaseModel
from .cv import *

model_factory = {
    "resnet": ResNet,
    "resnet18": ResNet,
    "resnet_mnist": ResNetMNIST,
    "mlp": MLP,
    "alexnet_cifar10": AlexNetCIFAR10,
    "cnn_cifar10": CNN2CIFAR10,
    "cnn2_cifar10": CNN2CIFAR10,
    "cnn4_cifar10": CNN4CIFAR10,
    "cnn_mnist": CNNMNIST,
    "densenet": DenseNet,
    "unet": UNet,
    "unet_mini": UNetMini,
}


def build_model(model_name: str):
    if model_name not in model_factory.keys():
        raise ValueError(f"Unknown model name: {model_name}")
    model = model_factory[model_name]
    return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\model\nn\resnet_mnist.py
# @Time    :   2022-05-03 16:07:27
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

import torch
import torch.nn as nn

from torchvision import models

from ..base_model import BaseModel


class ResNetMNIST(BaseModel):
    def __init__(self, args, model_time=None, model_version=0):
        super().__init__(args, model_time, model_version)
        self.input_size = (args.input_c, args.image_size, args.image_size)
        self.num_classes = args.num_classes
        self.conv = nn.Conv2d(1, 3, kernel_size=1)
        self.net = models.resnet18(pretrained=args.model_pretrained)
        self.net.fc = nn.Linear(512, args.num_classes)

    def forward(self, x):
        return self.net(self.conv(x))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\core\coordinator\base_coordinator.py
# @Time    :   2022-05-03 15:41:08
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

from abc import ABC, abstractmethod
from fedhf.api import Logger
from fedhf.core import build_server, build_client
from fedhf.component import build_sampler
from fedhf.dataset import ClientDataset, build_dataset


class AbsCoordinator(ABC):
    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    def prepare(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def main(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def finish(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def run(self) -> None:
        raise NotImplementedError


class SimulatedBaseCoordinator(AbsCoordinator):
    def __init__(self, args) -> None:
        super().__init__()
        self.args = args
        self.logger = Logger(self.args)

    def prepare(self) -> None:
        self.dataset = build_dataset(self.args.dataset)(self.args)
        self.sampler = build_sampler(self.args.sampler)(self.args)

        if self.args.test:
            # reduce data for test
            self.data = [
                ClientDataset(
                    self.dataset.trainset,
                    range(i * self.args.batch_size, (i + 1) * self.args.batch_size),
                )
                for i in range(self.args.num_clients)
            ]
        else:
            self.data = self.sampler.sample(self.dataset.trainset)

        self.client_list = [i for i in range(self.args.num_clients)]
        self.server = build_server(self.args.deploy_mode)(self.args)

    def main(self) -> None:
        pass

    def finish(self) -> None:
        self.server.model.save()

        try:
            if self.args.evaluate_on_client:
                self.logger.info("Evaluate on client")
                for client_id in self.client_list:
                    client = build_client(self.args.deploy_mode)(self.args, client_id)
                    result = client.evaluate(
                        data=self.data[client_id], model=self.server.model
                    )
                    self.logger.info(f"Client {client_id} result: {result}")

            result = self.server.evaluate(self.dataset.testset)
            self.logger.info(f"Server result: {result}")
            self.logger.info(
                f"Final server model version: {self.server.model.get_model_version()}"
            )
        except KeyboardInterrupt:
            self.logger.info(f"Interrupted by user.")

        self.logger.info(f"All finished.")

    def run(self) -> None:
        self.prepare()
        self.main()
        self.finish()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\api\type\__init__.py
# @Time    :   2022-02-13 14:25:31
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

from .const import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\component\sampler\noniid_sampler.py
# @Time    :   2022-05-03 16:00:43
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

import numpy as np

from fedhf.dataset import ClientDataset
from .base_sampler import BaseSampler


class NonIIDSampler(BaseSampler):
    def __init__(self, args) -> None:
        super(NonIIDSampler, self).__init__(args)

    def sample(self, train_dataset, test_dataset=None):
        """[summary]

        Reference: https://github.com/jeremy313/non-iid-dataset-for-personalized-federated-learning

        Args:
            dataset ([type]): [description]
            test_dataset ([type], optional): [description]. Defaults to None.

        Returns:
            [type]: [description]
        """
        num_items_train = int(len(train_dataset) / self.args.num_clients)
        sampler_num_samples = (
            self.args.sampler_num_samples
            if self.args.get("sampler_num_samples") is not None
            else 1
        )
        sampler_num_samples = int(sampler_num_samples)
        sampler_num_classes = (
            self.args.sampler_num_classes
            if self.args.get("sampler_num_classes") is not None
            else 1
        )
        sampler_num_classes = int(sampler_num_classes)
        num_shards_train = len(train_dataset) // sampler_num_samples

        if test_dataset is not None:
            num_items_test = len(test_dataset) // self.args.num_classes

        client_data_dict = {}

        idx_shard = [i for i in range(num_shards_train)]

        idxs = np.arange(self.args.num_clients * num_items_train, dtype=np.int32)
        labels = np.array(train_dataset.targets)

        if test_dataset is not None:
            idxs_test = np.arange(len(test_dataset), dtype=np.int32)
            labels_test = np.array(test_dataset.targets)

        idx_label = np.vstack((idxs, labels))
        idx_label = idx_label[:, idx_label[1, :].argsort()]
        idxs = idx_label[0, :]
        labels = idx_label[1, :]

        if test_dataset is not None:
            idx_label_test = np.vstack((idxs_test, labels_test))
            idxs_labels_test = idx_label_test[:, idx_label_test[1, :].argsort()]
            idxs_test = idxs_labels_test[0, :]
            labels_test = idxs_labels_test[1, :]

        client_data_train = {i: np.array([]) for i in range(self.args.num_clients)}
        client_data_test = {i: np.array([]) for i in range(self.args.num_clients)}

        unbalance_rate = (
            1.0
            if self.args.get("sampler_unbalance_rate") is None
            else self.args.sampler_unbalance_rate
        )
        unbalance_rate = float(unbalance_rate)

        for i in range(self.args.num_clients):
            rand_set = set(
                np.random.choice(idx_shard, sampler_num_classes, replace=False)
            )
            idx_shard = list(set(idx_shard) - rand_set)

            unbalance_flag = 0
            client_data_train[i] = np.array([], dtype=np.int32)
            client_label = np.array([])

            for rand_item in rand_set:
                if unbalance_flag == 0:
                    client_data_train[i] = np.concatenate(
                        (
                            client_data_train[i],
                            idxs[
                                sampler_num_samples
                                * rand_item : int(sampler_num_samples * (rand_item + 1))
                            ],
                        ),
                        axis=0,
                    )
                    client_label = np.concatenate(
                        (
                            client_label,
                            labels[
                                sampler_num_samples
                                * rand_item : int(sampler_num_samples * (rand_item + 1))
                            ],
                        ),
                        axis=0,
                    )
                else:
                    client_data_train[i] = np.concatenate(
                        (
                            client_data_train[i],
                            idxs[
                                sampler_num_samples
                                * rand_item : int(
                                    sampler_num_samples * (rand_item + unbalance_rate)
                                )
                            ],
                        ),
                        axis=0,
                    )
                    client_label = np.concatenate(
                        (
                            client_label,
                            labels[
                                sampler_num_samples
                                * rand_item : int(
                                    sampler_num_samples * (rand_item + unbalance_rate)
                                )
                            ],
                        ),
                        axis=0,
                    )
                unbalance_flag = 1
            client_label_set = set(client_label)

            if test_dataset is not None:
                for label in client_label_set:
                    client_data_test[i] = np.concatenate(
                        (
                            client_data_test[i],
                            idxs_test[
                                int(label)
                                * num_items_test : int(label + 1)
                                * num_items_test
                            ],
                        ),
                        axis=0,
                    )

        if test_dataset is None:
            return [
                ClientDataset(train_dataset, list(client_data_train[i]))
                for i in range(self.args.num_clients)
            ]
        else:
            return [
                (
                    ClientDataset(train_dataset, list(client_data_train[i])),
                    ClientDataset(test_dataset, list(client_data_test[i])),
                )
                for i in range(self.args.num_clients)
            ]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\component\sampler\random_sampler.py
# @Time    :   2022-05-03 16:00:48
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

import numpy as np

from fedhf.dataset import ClientDataset
from .base_sampler import BaseSampler


class RandomSampler(BaseSampler):
    def __init__(self, args) -> None:
        super(RandomSampler, self).__init__(args)

    def sample(self, dataset):
        num_items = int(len(dataset) / self.args.num_clients)
        client_data_dict, all_idxs = {}, [i for i in range(len(dataset))]
        for i in range(self.args.num_clients):
            client_data_dict[i] = list(
                np.random.choice(all_idxs, num_items, replace=False)
            )
            all_idxs = list(set(all_idxs) - set(client_data_dict[i]))

        return [
            ClientDataset(dataset, client_data_dict[i])
            for i in range(self.args.num_clients)
        ]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\__init__.py
# @Time    :   2022-02-13 14:32:13
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

# import utils api
from .api import *
from .algor import *
from .core import register, register_all


def init(*args, **kwargs):
    args = Config().parse_cfg(*args, **kwargs)
    args = init_algor(args)
    return args


def run(args=None):
    if args is None:
        args = init()

    from .core import build_coordinator

    coordinator_type = args.get("coordinator", f"{args.deploy_mode}_{args.scheme}")
    coordinator = build_coordinator(coordinator_type)(args)
    coordinator.run()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\core\server\simulated_server.py
# @Time    :   2022-05-03 03:34:11
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

from copy import deepcopy
import torch
import torch.nn as nn

from fedhf.api import Serializer, Deserializer

from .base_server import BaseServer


class SimulatedServer(BaseServer):
    def __init__(self, args) -> None:
        super(SimulatedServer, self).__init__(args)

    def update(self, model: nn.Module, **kwargs):
        # self.logger.info(f'Update model with {kwargs}')

        if self.model.get_model_version() == 0:
            self.model = deepcopy(model)

        result = self.aggregator.agg(
            Serializer.serialize_model(self.model),
            Serializer.serialize_model(model),
            **kwargs,
        )

        if not result:
            self.logger.info("It's not time to update.")
            return
        # print(self.model.get_model_version(), model.get_model_version())
        Deserializer.deserialize_model(self.model, result["param"])

        # self.model = self.encryptor.encrypt_model(self.model)

        self.model.set_model_version(result["model_version"])
        self.model.set_model_time(result["model_time"])
        # print(result['model_version'], result['model_time'])
        self.logger.info(
            f'get model version {result["model_version"]} at time {result["model_time"]}'
        )
        return



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\api\opt\opt.py
# @Time    :   2022-05-03 15:59:07
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

import argparse
import os
import yaml

import torch
import torch.nn as nn
import numpy as np


class opts(object):
    def __init__(self):
        self.parser = argparse.ArgumentParser()

        # high priority setting
        self.parser.add_argument("--cfg", type=str, default="", help="config file path")

        # basic experiment setting
        self.parser.add_argument(
            "--exp_name", default="exp", help="name of the experiment."
        )
        self.parser.add_argument(
            "--exist_ok", action="store_true", help="allow to overwrite existing files."
        )
        self.parser.add_argument(
            "--prj_name", default=None, help="name of the project."
        )
        self.parser.add_argument(
            "--deploy_mode",
            default="simulated",
            help="type of deployment. [ simulated, standalone, distributed ]",
        )
        self.parser.add_argument(
            "--scheme", default="sync", help="type of deployment. [ async, sync ]"
        )

        # system setting
        self.parser.add_argument(
            "--gpus", default="-1", help="-1 for cpu, use comma for multiple gpus"
        )
        self.parser.add_argument(
            "--num_workers",
            type=int,
            default=0,
            help="dataloader threads. 0 for single-thread.",
        )

        # random setting
        self.parser.add_argument("--seed", type=int, default=233, help="random seed")

        # log setting
        self.parser.add_argument(
            "--use_wandb", action="store_true", help="using wandb to store result"
        )
        self.parser.add_argument(
            "--wandb_log_client", action="store_true", help="log train on client or not"
        )
        self.parser.add_argument(
            "--wandb_reinit", action="store_true", help="reinit wandb"
        )
        self.parser.add_argument(
            "--log_name", default="logger", type=str, help="logger name"
        )
        self.parser.add_argument(
            "--log_dir", type=str, default=None, help="where to save the log."
        )
        self.parser.add_argument(
            "--log_file", default=None, type=str, help="where to save log"
        )
        self.parser.add_argument(
            "--log_metric", default=None, type=str, help="where to save metric"
        )
        self.parser.add_argument(
            "--log_level",
            default="debug",
            type=str,
            help="log level, it could be in [ error | warning | info | debug ]",
        )
        self.parser.add_argument(
            "--eval_interval", type=int, default=5, help="evaluation interval"
        )
        self.parser.add_argument(
            "--chkp_interval",
            type=int,
            default=50,
            help="when to save the model and result to disk.",
        )
        self.parser.add_argument(
            "--save_dir",
            type=str,
            default="./runs",
            help="where to save the result to disk.",
        )
        self.parser.add_argument(
            "--weights_dir",
            type=str,
            default=None,
            help="where to save the model weights to disk.",
        )

        # model setting
        self.parser.add_argument(
            "--model", type=str, default="cnn_mnist", help="model name."
        )
        self.parser.add_argument(
            "--model_pretrained",
            action="store_false",
            help="load pretrained model or not",
        )
        self.parser.add_argument(
            "--model_dir",
            default=os.path.join(".", "model"),
            type=str,
            help="path to download model",
        )  # Never used
        self.parser.add_argument(
            "--resume",
            default=None,
            type=str,
            help="resume an experiment. "
            "Reloaded the optimizer parameter and "
            "set load_model to args.name.pth "
            "in the save dir if load_model is empty.",
        )

        # data setting
        self.parser.add_argument(
            "--dataset",
            default="mnist",
            help="see fedhf/dataset for available datasets",
        )
        self.parser.add_argument("--resize", action="store_true", help="resize or not")
        self.parser.add_argument(
            "--data_dir", default=os.path.join(".", "dataset"), help="dataset directory"
        )
        self.parser.add_argument("--input_c", type=int, default=1, help="input channel")
        self.parser.add_argument(
            "--image_size", type=int, default=224, help="image_size"
        )
        self.parser.add_argument(
            "--output_c", type=int, default=1, help="output channel"
        )
        self.parser.add_argument(
            "--num_classes", type=int, default=10, help="number of classes"
        )

        self.parser.add_argument(
            "--trainer", type=str, default="default_trainer", help="trainer."
        )
        self.parser.add_argument(
            "--evaluator", type=str, default="default_evaluator", help="evaluator."
        )
        self.parser.add_argument("--optim", type=str, default="adam", help="optimizer.")
        self.parser.add_argument(
            "--momentum", type=float, default=0.75, help="momentum."
        )
        self.parser.add_argument(
            "--weight_decay", type=float, default=0.001, help="weight decay."
        )

        self.parser.add_argument(
            "--lr", type=float, default=1e-2, help="learning rate."
        )
        self.parser.add_argument(
            "--lr_scheduler", type=str, default=None, help="lr scheduler."
        )
        self.parser.add_argument("--lr_step", type=int, default=30, help="lr step.")

        self.parser.add_argument(
            "--loss", type=str, default="ce", help="loss function."
        )
        self.parser.add_argument(
            "--evaluate_on_client", action="store_true", help="evaluate on client"
        )

        # federated setting
        self.parser.add_argument(
            "--num_clients", type=int, default=100, help="clients number."
        )
        self.parser.add_argument(
            "--num_clients_per_round",
            type=int,
            default=None,
            help="clients number per round.",
        )
        self.parser.add_argument(
            "--num_epochs", type=int, default=3, help="training epochs."
        )
        self.parser.add_argument("--batch_size", type=int, default=8, help="batch size")
        self.parser.add_argument(
            "--num_rounds", type=int, default=5, help="server round."
        )

        self.parser.add_argument(
            "--sampler", type=str, default="random", help="data sample strategy"
        )

        self.parser.add_argument(
            "--selector", type=str, default="random", help="client select strategy"
        )
        self.parser.add_argument(
            "--select_ratio", type=float, default=0.5, help="select ratio"
        )
        self.parser.add_argument(
            "--algor", type=str, default="fedavg", help="federated algorithm"
        )

        # test setting
        self.parser.add_argument("--test", action="store_true", help="test mode")

    def parse(self, args=""):
        if args == "":
            opt = self.parser.parse_args()
        else:
            opt = self.parser.parse_args(args)

        return opt



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\algor\sync\fedavg\fedavg_aggregator.py
# @Time    :   2022-07-15 13:17:42
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0


import time
import torch
import torch.nn as nn

from ....component.aggregator.sync_aggregator import SyncAggregator


class FedAvgAggregator(SyncAggregator):
    def __init__(self, args) -> None:
        super(FedAvgAggregator, self).__init__(args)

    def agg(self, server_param: torch.Tensor, client_param: torch.Tensor, **kwargs):
        self._model_cached.append(client_param)
        self._model_counter += 1

        if "weight" not in kwargs.keys():
            kwargs["weight"] = 1 / self.num_clients_per_round

        self._model_weight[self._model_counter - 1] = kwargs["weight"]

        if not self._check_agg():
            return

        self.logger.info("Aggregate models")

        new_param = torch.zeros_like(server_param)
        for i in range(self.num_clients_per_round):
            new_param += self._model_weight[i] * self._model_cached[i]

        self._model_cached = []
        self._model_counter = 0
        self._model_weight = [
            1 / self.num_clients_per_round for i in range(self.num_clients_per_round)
        ]

        result = {
            "param": new_param,
            "model_version": kwargs["server_model_version"] + 1
            if "server_model_version" in kwargs.keys()
            else 0,
            "model_time": time.time(),
        }
        return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\component\evaluator\__init__.py
# @Time    :   2022-05-03 16:00:21
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

__all__ = ["build_evaluator", "evaluator_factory", "BaseEvaluator", "DefaultEvaluator"]

from .base_evaluator import BaseEvaluator
from .default_evaluator import DefaultEvaluator


evaluator_factory = {
    "base_evaluator": BaseEvaluator,
    "default_evaluator": DefaultEvaluator,
}


def build_evaluator(name):
    if name not in evaluator_factory.keys():
        raise ValueError(f"Unknown evaluator name: {name}")
    return evaluator_factory[name]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\component\encryptor\__init__.py
# @Time    :   2022-05-02 22:57:05
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

from .base_encryptor import AbsEncryptor, BaseEncryptor
from .dp_encryptor import DPEncryptor

encryptor_factory = {
    "none": BaseEncryptor,
    "base": BaseEncryptor,
    "dp": DPEncryptor,
}


def build_encryptor(encryptor_name):
    """
    Build the crypor mechanism.
    :param encryptor_name: the name of the crypor mechanism
    :return: the crypor mechanism
    """
    if encryptor_name not in encryptor_factory:
        raise ValueError("Unknown crypor mechanism: {}".format(encryptor_name))
    return encryptor_factory[encryptor_name]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\core\coordinator\simulated_sync_coordinator.py
# @Time    :   2022-05-03 16:02:43
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

from copy import deepcopy

from fedhf.core import build_client

from .base_coordinator import SimulatedBaseCoordinator


class SimulatedSyncCoordinator(SimulatedBaseCoordinator):
    """Simulated Coordinator
    In simulated scheme, the data and model belong to coordinator and there is no need communicator.
    Also, there is no need to instantiate every client.
    """

    def __init__(self, args) -> None:
        super(SimulatedSyncCoordinator, self).__init__(args)

    def prepare(self) -> None:
        super(SimulatedSyncCoordinator, self).prepare()

        assert self.args.deploy_mode == "simulated"

    def main(self) -> None:
        try:
            for i in range(self.args.num_rounds):
                selected_client = self.server.select(self.client_list)

                self.logger.info(f"Round {i} selected client: {selected_client}")

                for client_id in selected_client:
                    model = deepcopy(self.server.model)
                    client = build_client(self.args.deploy_mode)(
                        self.args, client_id, data_size=len(self.data[client_id])
                    )
                    model, result = client.train(self.data[client_id], model)
                    self.server.update(
                        model,
                        server_model_version=self.server.model.get_model_version(),
                        client_id=client_id,
                    )

                result = self.server.evaluate(self.dataset.testset)
                self.logger.info(f"Server result: {result}")

                if self.server.model.get_model_version() % self.args.chkp_interval == 0:
                    self.server.model.save(
                        f"{self.args.exp_name}-{self.server.model.get_model_version()}.pth"
                    )

            self.logger.info(f"All rounds finished.")

        except KeyboardInterrupt:
            self.server.model.save()
            self.logger.info(f"Interrupted by user.")

    def finish(self) -> None:
        super(SimulatedSyncCoordinator, self).finish()

    def run(self) -> None:
        self.prepare()
        self.main()
        self.finish()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\core\client\base_client.py
# @Time    :   2022-05-03 15:55:52
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

from abc import ABC, abstractmethod

from fedhf.api import Logger
from fedhf.component import build_evaluator, build_trainer, build_encryptor


class AbsClient(ABC):
    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    def train(self):
        raise NotImplementedError

    def evalute(self):
        raise NotImplementedError


class BaseClient(AbsClient):
    def __init__(self, args, client_id) -> None:
        self.args = args
        self.client_id = client_id

        self.trainer = build_trainer(self.args.trainer)(self.args)
        self.evaluator = build_evaluator(self.args.evaluator)(self.args)
        self.encryptor = None
        if self.args.get("encryptor"):
            self.encryptor = build_encryptor(self.args.encryptor)(self.args)

        self.logger = Logger(self.args)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\api\__init__.py
# @Time    :   2022-05-03 03:32:25
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

from .cfg import *
from .logger import *
from .serial import *

from .type import *
from .dpm import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\model\__init__.py
# @Time    :   2022-05-03 16:07:56
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

from .criterion import *
from .nn import *
from .optimizer import *
from .scheduler import *
from .nn import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\component\aggregator\base_aggregator.py
# @Time    :   2022-05-03 15:59:51
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

from abc import ABC, abstractmethod
from fedhf.api import Logger


class AbsAggregator(ABC):
    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    def agg(self):
        raise NotImplementedError

    @abstractmethod
    def _check_agg(self):
        raise NotImplementedError


class BaseAggregator(AbsAggregator):
    def __init__(self, args) -> None:
        self.args = args
        self.logger = Logger(self.args)

    def agg(self):
        pass

    def _check_agg(self):
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\dataset\random.py
# @Time    :   2022-05-03 03:33:42
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.utils.data import Dataset, DataLoader


class RandomDataset(Dataset):
    """
    Generate random data.
    """

    def __init__(self, args, length, data_shape, num_classes=None):
        self.args = args
        self.length = length
        self.data_shape = data_shape
        self.num_classes = num_classes if num_classes is not None else args.num_classes
        self.num_classes = self.num_classes if self.num_classes is not None else 10

        self.data = torch.randn(length, *data_shape)
        self.label = torch.randint(0, self.num_classes, (length,))

    def __getitem__(self, index):
        return self.data[index], self.label[index]

    def __len__(self):
        return self.length



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\algor\async_\fedasync\fedasync_trainer.py
# @Time    :   2022-07-15 13:17:29
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0


import time
from copy import deepcopy

from tqdm import tqdm
from ....component.trainer.base_trainer import BaseTrainer


class FedAsyncTrainer(BaseTrainer):
    def __init__(self, args) -> None:
        super(FedAsyncTrainer, self).__init__(args)

        self.rho = args.fedasync.get("rho", 0.005)
        self.args.fedasync.update({"rho": self.rho})

    def train(
        self,
        dataloader,
        model,
        num_epochs,
        client_id=None,
        gpus=[],
        device="cpu",
        encryptor=None,
    ):
        if len(gpus) > 1:
            pass
        else:
            pass

        model_ = deepcopy(model)
        model = model.to(device)
        model_ = model_.to(device)
        if self.args.optim == "sgd":
            optim = self.optim(
                params=model.parameters(),
                lr=self.args.lr,
                momentum=self.args.momentum,
                weight_decay=self.args.weight_decay,
            )
        else:
            optim = self.optim(params=model.parameters(), lr=self.args.lr)
        crit = self.crit()
        if self.lr_scheduler is not None:
            lr_scheduler = self.lr_scheduler(optim, self.args.lr_step)

        self.logger.info(f"Start training on {client_id}")

        train_loss = []
        pbar = tqdm(total=num_epochs * len(dataloader))
        model.train()
        for epoch in range(num_epochs):
            losses = []
            pbar.set_description(
                f"Client:{client_id} Training on Epoch {epoch+1}/{num_epochs} Loss: {0 if len(train_loss)==0 else train_loss[-1]:.5f}"
            )
            for idx, (inputs, labels) in enumerate(dataloader):
                inputs = inputs.to(device)
                labels = labels.to(device)

                outputs = model(inputs)

                # self.logger.info(f'labels: {labels}, outputs: {outputs}')

                l2_reg = self._calc_l2_reg(model_, model)

                loss = crit(outputs, labels) + l2_reg * self.rho / 2

                # self.logger.info(
                #    f'ce: {crit(outputs, labels)}, l2_reg: {l2_reg}, loss: {loss.item()}')

                optim.zero_grad()

                if encryptor is not None:
                    encryptor.clip_grad(model)

                loss.backward()
                optim.step()

                losses.append(loss.item())
                pbar.update(1)

            train_loss.append(sum(losses) / len(losses))
            if self.lr_scheduler is not None:
                lr_scheduler.step()
            # self.logger.info(
            #    f'Client:{client_id} Epoch:{epoch+1}/{num_epochs} Loss:{train_loss[-1]}'
            # )

        time.sleep(0.3)
        self.logger.info(f"Client:{client_id} Train Loss:{train_loss}")

        if self.args.use_wandb and self.args.wandb_log_client:
            import wandb

            data = [[x, y] for (x, y) in zip(range(1, num_epochs + 1), train_loss)]
            table = wandb.Table(data=data, columns=["epoch", "train_loss"])
            self.logger.to_wandb(
                {
                    f"train at client {client_id} model_version {model.get_model_version()}": wandb.plot.line(
                        table,
                        "epoch",
                        "train_loss",
                        title=f"train loss at client {client_id}",
                    )
                }
            )

        return {"train_loss": train_loss, "model": model}

    def _calc_l2_reg(self, global_model, model):
        l2_reg = 0
        for p1, p2 in zip(global_model.parameters(), model.parameters()):
            l2_reg += (p1 - p2).norm(2)
        return l2_reg



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\model\nn\base_model.py
# @Time    :   2022-05-03 16:07:05
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

from copy import deepcopy
import time
import os

import torch
import torch.nn as nn


class BaseModel(nn.Module):
    def __init__(self, args, model_time=None, model_version=0):
        super(BaseModel, self).__init__()
        self.args = args
        self.model_time = model_time if model_time else time.time()
        self.model_version = model_version

    def get_model_time(self):
        return self.model_time

    def set_model_time(self, model_time):
        self.model_time = model_time

    def get_model_version(self):
        return self.model_version

    def set_model_version(self, model_version):
        self.model_version = model_version

    def save(self, path: str = None) -> None:
        if path is None:
            path = os.path.join(self.args.save_dir, f"{self.args.exp_name}.pth")
        torch.save(
            obj={
                "model_version": self.model_version,
                "model_time": self.model_time,
                "state_dict": self.state_dict(),
            },
            f=path,
        )

    def load(self, path: str = None) -> None:
        if path is None:
            path = os.path.join(self.args.save_dir, f"{self.args.exp_name}.pth")
        checkpoint = torch.load(path)
        self.model_version = checkpoint["model_version"]
        self.model_time = checkpoint["model_time"]
        self.load_state_dict(checkpoint["state_dict"])

    def add_default_args(self, args) -> None:
        self.args.merge(args, overwrite=False)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\component\sampler\base_sampler.py
# @Time    :   2022-05-03 16:00:39
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

from abc import ABC


class BaseSampler(ABC):
    def __init__(self, args) -> None:
        self.args = args

    def sample(self):
        raise NotImplementedError



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\algor\__init__.py
# @Time    :   2022-07-15 12:54:38
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

from typing import Any
from .base import build_algor
from fedhf.core.registor import register
from fedhf.api import Config


def init_algor(args) -> Any:
    # register components
    algor = build_algor(args.algor)
    temp_cfg = None
    if args.get(args.algor, None) is not None:
        temp_cfg = args.get(args.algor).deepcopy()

    # update parameters
    args.update(algor.default_params)
    if temp_cfg is not None:
        args[args.algor].update(temp_cfg)

    for component_name, component in algor.components.items():
        register(component_name, component)

    return args



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\core\client\__init__.py
# @Time    :   2022-05-03 16:02:14
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

__all__ = ["SimulatedClient", "build_client", "client_factory"]

from .simulated_client import SimulatedClient

client_factory = {"simulated": SimulatedClient}


def build_client(client_type):
    if client_type not in client_factory.keys():
        raise ValueError(
            "client_type {} not in {}".format(client_type, client_factory.keys())
        )
    client = client_factory[client_type]
    return client



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\dataset\mnist.py
# @Time    :   2022-05-03 16:06:35
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

from torchvision.datasets import MNIST
from torchvision.transforms import Compose, ToTensor, Normalize, Resize
import torch


class MNISTDataset(object):
    def __init__(self, args) -> None:
        super().__init__()

        self.args = args
        self.num_classes = 10

        if args.resize:
            self.transform = Compose(
                [Resize([args.image_size, args.image_size]), ToTensor()]
            )
        else:
            self.transform = Compose([ToTensor()])

        self.trainset = MNIST(
            root=args.data_dir, train=True, download=True, transform=self.transform
        )
        self.testset = MNIST(
            root=args.data_dir, train=False, download=True, transform=self.transform
        )

        self.trainset.num_classes = self.num_classes
        self.testset.num_classes = self.num_classes

    def get_train_norm(self) -> float:
        """Never used"""
        means = []
        stds = []
        for img, label in self.trainset:
            means.append(torch.mean(img))
            stds.append(torch.std(img))

        mean = torch.mean(torch.tensor(means))
        std = torch.mean(torch.tensor(stds))
        return {"mean": mean, "std": std}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\dataset\__init__.py
# @Time    :   2022-05-03 16:06:13
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

from .mnist import MNISTDataset
from .cifar10 import CIFAR10Dataset
from .random import RandomDataset
from .client_datasest import ClientDataset

dataset_factory = {
    "mnist": MNISTDataset,
    "cifar10": CIFAR10Dataset,
    "random": RandomDataset,
}


def build_dataset(dataset_name: str):
    if dataset_name not in dataset_factory.keys():
        raise NotImplementedError

    dataset = dataset_factory[dataset_name]

    return dataset


__all__ = [
    "build_dataset",
    "dataset_factory",
    "MNISTDataset",
    "CIFAR10Dataset",
    "RandomDataset",
    "ClientDataset",
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\dataset\client_datasest.py
# @Time    :   2022-05-03 16:06:28
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

from typing import List
from torch.utils.data import Dataset


class ClientDataset(Dataset):
    def __init__(self, dataset, data_dict: List) -> None:
        super().__init__()
        self.dataset = dataset
        self.data_dict = data_dict

        for attr in self.dataset.__dict__:
            if attr.startswith("__") or attr in self.__dict__:
                continue
            # print(attr)
            setattr(self, attr, getattr(self.dataset, attr))

    def __len__(self):
        return len(self.data_dict)

    def __getitem__(self, index):
        return self.dataset[self.data_dict[index]]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\api\serial\deserializer.py
# @Time    :   2022-05-03 15:59:24
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

import io
import torch


class Deserializer(object):
    """
    Deserializer
    """

    @staticmethod
    def deserialize_model(
        model: torch.nn.Module, serialized_parameters: torch.Tensor, mode="copy"
    ):
        cur_idx = 0
        for parameter in model.parameters():
            numel = parameter.data.numel()
            size = parameter.data.size()
            if mode == "copy":
                parameter.data.copy_(
                    serialized_parameters[cur_idx : cur_idx + numel].view(size)
                )
            elif mode == "add":
                parameter.data.add_(
                    serialized_parameters[cur_idx : cur_idx + numel].view(size)
                )
            else:
                raise ValueError("Unknown mode: {}".format(mode))
            cur_idx += numel



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\core\server\__init__.py
# @Time    :   2022-05-03 15:45:20
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

__all__ = ["SimulatedServer", "build_server", "server_factory"]

from .simulated_server import SimulatedServer

server_factory = {
    "simulated": SimulatedServer,
}


def build_server(server_type: str):
    if server_type not in server_factory:
        raise ValueError("Unknown server type: {}".format(server_type))
    server = server_factory[server_type]
    return server



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\component\selector\random_selector.py
# @Time    :   2022-05-03 16:01:05
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

import numpy as np

from .base_selector import BaseSelector


class RandomSelector(BaseSelector):
    def __init__(self, args) -> None:
        self.args = args

    def select(self, client_list: list) -> list:
        num_client_per_round = self.args.num_clients_per_round or int(
            self.args.num_clients * self.args.select_ratio
        )
        selected_clients = np.random.choice(
            client_list, num_client_per_round, replace=False
        )
        return selected_clients



from setuptools import setup, find_packages
from os import path as os_path

__version__ = "0.3.0"

this_directory = os_path.abspath(os_path.dirname(__file__))

setup(
    name="fedhf",
    version=__version__,
    keywords=[
        "federated learning",
        "deep learning",
        "pytorch",
        "asynchronous federated learning",
        "wandb",
    ],
    author="Bingjie Yan",
    author_email="bj.yan.pa@qq.com",
    maintainer="Bingjie Yan",
    maintainer_email="bj.yan.pa@qq.com",
    description="A Federated Learning Framework which is Heterogeneous and Flexible.",
    long_description=open(os_path.join(this_directory, "README.md")).read(),
    long_description_content_type="text/markdown",
    license="Apache-2.0 License",
    url="https://github.com/beiyuouo/fedhf",
    packages=find_packages(include=["fedhf", "fedhf.*", "LICENSE"]),
    package_data={"": ["*.yaml"]},
    install_requires=[
        "torch>=1.8.2",
        "torchvision>=0.9.2",
        "numpy",
        "tqdm",
        "wandb",
        "ezkfg",
        "pyyaml",
        "urllib3==1.25.3",
    ],
    extras_require={
        "dev": ["nox", "pytest", "mkdocs"],
        "test": ["pytest", "mkdocs"],
    },
    python_requires=">=3.6",
    classifiers=[
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Software Development",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
    ],
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\component\trainer\base_trainer.py
# @Time    :   2022-05-03 16:01:18
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

from abc import ABC, abstractmethod

from fedhf.api import Logger
from fedhf.model import build_criterion, build_optimizer, build_lr_scheduler


class AbsTrainer(ABC):
    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    def train(self):
        raise NotImplementedError


class BaseTrainer(AbsTrainer):
    def __init__(self, args):
        self.args = args
        self.optim = build_optimizer(self.args.optim)
        self.crit = build_criterion(self.args.loss)
        if self.args.lr_scheduler:
            self.lr_scheduler = build_lr_scheduler(self.args.lr_scheduler)
        else:
            self.lr_scheduler = None
        self.logger = Logger(self.args)

    def set_device(self, gpus, device):
        if len(gpus) > 1:
            pass
        else:
            pass

    def train(self):
        print("BaseTrainer")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\core\coordinator\__init__.py
# @Time    :   2022-05-03 16:02:27
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

__all__ = [
    "build_coordinator",
    "coordinator_factory",
    "SimulatedBaseCoordinator",
    "SimulatedSyncCoordinator",
    "SimulatedAsyncCoordinator",
]

from .base_coordinator import SimulatedBaseCoordinator

from .simulated_sync_coordinator import SimulatedSyncCoordinator
from .simulated_async_coordinator import SimulatedAsyncCoordinator

coordinator_factory = {
    "simulated_sync": SimulatedSyncCoordinator,
    "simulated_async": SimulatedAsyncCoordinator,
}


def build_coordinator(coordinator_type):
    if coordinator_type not in coordinator_factory:
        raise ValueError(f"Unknown coordinator type: {coordinator_type}")
    coordinator = coordinator_factory[coordinator_type]
    return coordinator



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\api\dpm\gaussian_noise.py
# @Time    :   2022-05-02 22:35:11
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

import numpy as np
import torch


def gaussian_noise(sensitivity, size, epsilon, **kwargs):
    """
    Generate Gaussian noise with the given sensitivity.
    :param sensitivity: the sensitivity of the privacy mechanism
    :param size: the size of the noise
    :param epsilon: the privacy parameter
    :param kwargs: other parameters
    :return: the generated noise
    """
    assert "delta" in kwargs, "delta is required"
    delta = kwargs["delta"]
    noise_scale = np.sqrt(2 * np.log(1.25 / delta)) * sensitivity / epsilon
    return np.random.normal(0, noise_scale, size)


def gaussian_clip(model: torch.nn.Module, clip: float):
    """
    Clip the model parameters.
    :param model: the model
    :param clip: the clipping bound
    :return: None
    """
    for k, v in model.named_parameters():
        v.grad /= max(1, v.grad.norm(2) / clip)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\component\trainer\__init__.py
# @Time    :   2022-05-03 16:01:12
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

__all__ = ["build_trainer", "trainer_factory", "BaseTrainer", "DefaultTrainer"]

from .base_trainer import BaseTrainer
from .default_trainer import DefaultTrainer


trainer_factory = {
    "base_trainer": BaseTrainer,
    "default_trainer": DefaultTrainer,
}


def build_trainer(trainer_type: str):
    if trainer_type not in trainer_factory.keys():
        raise ValueError(f"{trainer_type} is not a valid trainer name")

    return trainer_factory[trainer_type]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\component\evaluator\evaluator.py
# @Time    :   2022-05-03 16:00:29
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

from typing import Any, Optional, Union
import torch
import torch.nn as nn
from torch.utils.data import DataLoader
from tqdm import tqdm

from fedhf.model import build_criterion, build_optimizer
from .base_evaluator import BaseEvaluator


class DefaultEvaluator(BaseEvaluator):
    def __init__(self, args) -> None:
        super(DefaultEvaluator, self).__init__(args)

    def evaluate(
        self,
        dataloader: DataLoader,
        model: nn.Module,
        client_id: Union[int, str] = None,
        gpus: list = [],
        device="cpu",
    ) -> Any:
        if len(gpus) > 1:
            pass
        else:
            pass
        if not client_id:
            client_id = -1
        model = model.to(device)
        crit = self.crit()

        self.logger.info(f"Start evaluation on {client_id}")

        model.eval()
        losses = 0.0
        acc = 0.0
        for inputs, labels in tqdm(dataloader, desc=f"Test on client {client_id}"):
            inputs = inputs.to(device)
            labels = labels.to(device)

            outputs = model(inputs)
            loss = crit(outputs, labels)

            _, predicted = torch.max(outputs, 1)
            acc += torch.sum(predicted == labels).item()

            losses += loss.item()

        losses /= len(dataloader)
        # self.logger.info(f'Client {client_id} test loss: {losses:.4f}, acc: {acc}')
        acc /= len(dataloader.dataset)

        self.logger.info(f"Evaluation on {client_id} finished")

        if self.args.use_wandb:
            if client_id == -1:
                self.logger.to_wandb(
                    {
                        "acc on server": acc,
                        "loss on server": losses,
                        "epoch": model.get_model_version(),
                    }
                )

        return {"test_loss": losses, "test_acc": acc}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\component\aggregator\__init__.py
# @Time    :   2022-05-03 15:59:40
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

__all__ = [
    "build_aggregator",
    "SyncAggregator",
    "AsyncAggregator",
    "aggregator_factory",
    "BaseAggregator",
]

from .base_aggregator import BaseAggregator
from .sync_aggregator import SyncAggregator
from .async_aggregator import AsyncAggregator

aggregator_factory = {
    "base": BaseAggregator,
    "sync": SyncAggregator,
    "async": AsyncAggregator,
}


def build_aggregator(agg_name: str):
    if agg_name not in aggregator_factory.keys():
        raise ValueError(f"Unknown aggregator name: {agg_name}")
    agg = aggregator_factory[agg_name]
    return agg



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File    :   fedhf\component\selector\__init__.py
# @Time    :   2022-05-03 16:00:53
# @Author  :   Bingjie Yan
# @Email   :   bj.yan.pa@qq.com
# @License :   Apache License 2.0

__all__ = ["build_selector", "selector_factory", "RandomSelector"]

from .random_selector import RandomSelector

selector_factory = {
    "random": RandomSelector,
}


def build_selector(sele_name: str):
    if sele_name not in selector_factory.keys():
        raise ValueError(f"{sele_name} is not a valid selector name")

    selector = selector_factory[sele_name]

    return selector
